/*******************************************************************************
 * Copyright 2017 Aerospike, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
#pragma once

#include <aerospike/aerospike.h>
as_status send_info_to_tls_host(aerospike* as, as_error* err, const as_policy_info* info_policy,
							    const char* hostname, uint16_t port, const char* tls_name,
							    const char* request, char** response);
